#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 10 16:32:22 2017

@author: hielke

This module contains functions for processing the CARD database with webcrawling
"""

# This code is part of the software of iGEM TUDelft 2017 and governed by 
# its license.  Please see the LICENSE file that is also included.

# Pythonic modules
from collections import defaultdict
from urllib.request import urlopen

# Third party modules / You should make sure you download this
from bs4 import BeautifulSoup

# My own modules
from custom_functions import cd, logit, mkdir, grouper

@logit
def getSoup(url):
    "From a given url, returns the object containing the parsed html"
    print("Requesting: " + url)
    with urlopen(url) as res:
        soup = BeautifulSoup(res.read(), 'html5lib')
    return soup

  
def getSubTermsFromSoup(soup):
    "Collect the sub-Term(s) from the given soup"
    li = soup.find(class_='panel-body').table.tbody.find_all('tr')
    
    titles = [next(tr.td.children) for tr in li]
    
    subTermsSection = li[titles.index('Sub-Term(s)')]
    
    subTerms = subTermsSection.find_all('td')[1]
       
    sub_rel = [sub for sub in subTerms.find_all('a') if sub.get('href') is not None]
    
    sub_rel_grouped = grouper(sub_rel, 2)
    
    res = defaultdict(list) # creates dict with new entries initialized as lists
    for sub in sub_rel_grouped:
        res[sub[1].string].append((sub[0].string, sub[0]['href']))
    
    return res

def saveFasta(fastaDNA, webID):
    "from the fasta data, create a fastafile"
    fastafile = fastaDNA.contents[0]
    filename = '_'.join(fastafile.split('\r', 1)[0].split('|', 2)[0:2])[1:]
    filename += '_' + webID + '.fasta'
    
    with open(filename, 'w') as f:
        f.write(fastafile)

# =============================================================================
# global accesseble functions
# =============================================================================
        
def getSubTerms(entry, BASE):
    """
    Collect the sub-Term(s) from the given card entry
    
    * Default is 'antibiotic molecule'
    * Returns dict with keys as relation and list with the entries
    """
    
    soup = getSoup(BASE + entry[1])
    
    return getSubTermsFromSoup(soup)
    

def processGenes(entry, BASE):
    """
    processes a gene entry or an entry to multiple by producing the fasta file 
    of the DNA sequence
    """
    
    soup = getSoup(BASE + entry[1])
    
    fastaDNA = soup.find(id='fastaDNA')
    if fastaDNA is None:
        name = mkdir(entry)
        with cd(name):
            for otherEntry in getSubTermsFromSoup(soup)['is_a']:
                processGenes(otherEntry)
    else:
        saveFasta(fastaDNA, webID=entry[1].split('/')[2])
    

# =============================================================================
# Testing
# =============================================================================

if __name__ == '__main__':
    start = ('antibiotic molecule', '/ontology/36008')
    for gene in getSubTerms(getSubTerms(start)['is_a'][0])['confers_resistance_to']:
        processGenes(gene)
