#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 12 14:58:06 2017

@author: hielke

Just some usefull general functions. 

The first are available on all operating systems, but the second half requires
the use of *NIX operating system. (Like mac, or Linux, but Windows with cygwin
installed might also mentioned.) The functions of the second half are only used
in the scripts related to the meme program which by itself can only be installed 
on *NIX operating systems. 
"""

# This code is part of the software of iGEM TUDelft 2017 and governed by 
# its license.  Please see the LICENSE file that is also included.

import os
import time
import argparse
from contextlib import contextmanager
from itertools import zip_longest
from subprocess import PIPE, Popen

# =============================================================================
# Available on all operating systems (noarch).
# =============================================================================

def saveCat(*args):
    '''
    catenates all iterables in the input, and leaves out the None types
    if all are None, return None
    '''
    res = None
    isNone = True
    for arg in args:
        if isNone and arg:
            res = []
            isNone = False
        if not isNone:
            arg = arg or []
            res += arg

    return res
        
def mkdir(entry):
    "Creates a directory from an entry, checks also if already exists"
    name = entry[0].strip().replace(' ', '_') + '_' + entry[1].split('/')[2]
    if not os.path.exists(name):
        os.mkdir(name)
    return name
    
@contextmanager
def cd(newdir, verbose=True):
    "performs a change of directory, but with 'with' context manangement"
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try: 
        yield
    except Exception as e:
        print(e)
        print("... when working in " + newdir)
    finally:
        if verbose:
            print("Done working in %r.\n" % os.getcwd())
        os.chdir(prevdir)

def logit(f):
    "A decorator for logging all the important information"
    def wrapper(*args, **kwargs):
        wrapper.counter += 1
        ts = time.time()
        result = f(*args, **kwargs)
        te = time.time()
        elapsed = te - ts
        wrapper.total += elapsed
        string = "Calling %r for the %d time.\n" % (f.__name__, wrapper.counter)
        string += "This time %r took %2.2f seconds.\n" % (f.__name__, elapsed)
        string += "%r has ran for a total of %5.2f seconds\n" % \
                (f.__name__, wrapper.total)
        string += "with an average time of %2.2f seconds.\n" % \
                (wrapper.total / wrapper.counter)
        print(string)
        return result
    wrapper.total = 0 
    wrapper.counter = 0
    return wrapper

class SmartFormatter(argparse.HelpFormatter):

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()  
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)

def grouper(iterable, n, fillvalue=None):
    """
    Collect data into fixed-length chunks or blocks
    
    >>> grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


# =============================================================================
# Functions that require a *NIX based OS:
# =============================================================================

def sh(cmd, verbose=True):
    """Returns the stdout of the shell command as an iterator over the lines.
    The lines are strip()-ed. 
    """
    if verbose:
        print("[INFO]: executing: " + cmd)
    out, err = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True).communicate()
    if err:
        print("[ERROR]: while executing: " + cmd)
        print(err.decode('ascii').strip())
    return out.decode('ascii').strip()

def fileCat(extension, verbose=True, Windows=False):
    """
    In the current folder collect all the files with the extension 'extension' 
    recursively and catenate them. It will not take into account files that 
    already have the word multiple prefixed to their extension. The last 
    condition makes sure this function can be repeated without collecting duplicates.
    As a bonus converts the newline character to standard \n
    
    Creates file temp.gathered.multiple.extension with unaltered newline characters
    and file gathered.multiple.extension with altered newline characters
    
    NB. Windows might experience problems with this. so if you want you can 
    pass windows=True in the function. 
    
    NB2. Requires sed version 4.2.2, because of the -z option. 
    Usually you will find it any modern *NIX based OS. 
    """
    assert extension[0] == '.', \
            "extension should start with a '.', not %s" % extension[0]
    temp_filename = "temp.gathered.multiple" + extension
    gathered_filename = "gathered.multiple" + extension
    sh('> ' + temp_filename, verbose) # empty file contents
    for l in sh('find -name "*%s"' % extension, verbose).splitlines():
        if "multiple" not in l:
            sh('(cat "%s"; echo) >> %s' % (l, temp_filename), verbose)
    if Windows:
        sh("sed -z 's/\\n/\\r\\n/g; s/\\r/\\r\\n/g < " 
               + temp_filename + ' > ' + gathered_filename, verbose)
    else:
        sh("sed -z 's/\\r\\n/\\n/g; s/\\r/\\n/g' < " 
               + temp_filename + ' > ' + gathered_filename, verbose)
        

def nonEmptyDir() -> bool:
    """
    Lists all nonempty directories recursively
    NB. Returns a filter object (an iterator)
    
    Requires *NIX based OS.
    """
    all_dir = sh("find . -type d").splitlines()
    empty_dir = list(sh("find . -type d -empty")).splitlines()
    nonempty_dir = filter(lambda dr: dr not in empty_dir, all_dir)
    return nonempty_dir

# =============================================================================
# Testing
# =============================================================================

if __name__ == '__main__':
    args = (None, [3,4], None, [2,3,4], ['hoi'])
    print(saveCat(*args))
    print(saveCat(None, None))
    
    subdir = 'subdir'
    if not os.path.exists('subdir'):
        os.mkdir(subdir)
    with cd(subdir):
        print(os.getcwd())
    os.rmdir('subdir')
    assert list(grouper(range(7), 3, 'X')) == [(0,1,2), (3,4,5), (6,'X','X')]
    assert sh('cd `echo $HOME`; cd ..; pwd')== '/home'
    print(sh('ls'))
    with cd('test/tetracycline_35968'):
        fileCat('.fasta')
