#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 21 22:36:04 2017

@author: hielke

This module is related to function for sending the correct jobs to the meme utility, 
and for post-processing the results from it. 
"""

# This code is part of the software of iGEM TUDelft 2017 and governed by 
# its license.  Please see the LICENSE file that is also included.

from motif_reader import read
from Bio.SeqUtils import GC
from custom_functions import sh

# =============================================================================
# FUNCTIONS FOR STARTING THE CORRECT JOBS TO SEND TO MEME
# =============================================================================

def memeProcess(f:str, i:int, n:int) -> None:
    """
    This executes the meme program. It spawns as a new process, 
    not as a new thread.
    """
    sh(memeCmd(f, i, n))
    

def memeOnFile(f:str):
    """
    This function is a factory. It returns a function. The returned function 
    it the memeProcess function, but with the file already set.
    """
    def memeOnNumber(i, n):
        return memeProcess(f, i, n)
    return memeOnNumber

def memeCmd(file:str, sites:int, motifs:int, width:int=24) -> str:
    """
    Create a command for the terminal running the meme process on 'file' with 
    a maximum number of sites to check and a number of motifs to find. 
    
    The number of motifs largely determines the time the program will run. 
    
    The number of sites will largely determine how good the results will be. 
    If you attempt to cover more sites, the result will have more SNPs in them.
    
    Optional, you can specify the width of the motif (default is 24).
    """
    
    return ('meme %s -o meme_output_%i -dna -revcomp -w %i -nmotifs' 
            + '%i -maxsites %i -maxsize 1000000000' 
                            % (file, sites, width, motifs, sites))

# =============================================================================
# FUNCTIONS FOR POST-PROCESSING YOUR RESULTS
# =============================================================================

def only_one(motif, loc):
    """
    Returns false if in the counts array there is only one base. 
    (So a 100 % score.)
    
    (This might be a little strict though, a motif could contain valuable data, 
    but still got thrown away while it might be just one sequence that ruins
    the 100 % score.)
    """
    not_the_highest_scored = sorted(list(motif.counts[:,loc].values()))[:-1]
    # return true if not_the_highest score is all zero
    return not any(not_the_highest_scored) 

def wrong(motif):
    "Returns the amount of SNPs and the place it is in." 
    oo = [only_one(motif, loc) for loc in range(len(motif))]
    places = [i for i, x in enumerate(oo) if not x]
    som = sum([not i for i in oo])
    return (som, places)

def passed(motif):
    """
    Returns true if the motif passes the requirements for Cas13a.
    
    * The is GC between 40 and 60 percent.
    * No more than 2 SNPs.
    * If there are 2 SNPs, don't allow any between position 9 and 16, \
    that is the place of the seed region.
    """
    if GC(motif.consensus) < 40 or 60 < GC(motif.consensus):
        return False
    wr = wrong(motif)
    if wr[0] < 2:
        return True
    if wr[0] > 2:
        return False
    if wr[0] == 2:
        # AND if there is any wrong base between 9 and 16, return false
                # otherwise return true
        return not any([9<i and i<16 for i in wr[1]])

if __name__ == '__main__':
    with open("meme.txt") as handle:
        record = read(handle)
    motif = record[2]        
    print(motif.degenerate_consensus, ' : ', passed(motif))
    print()
    print([instance.sequence_name for instance in motif.instances])
    
    print(record.sequences)
    


