#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 24 15:26:37 2017

@author: hielke / hielkewalinga@gmail.com
"""

# Third party plotting library / You have to download/install this
from matplotlib import pyplot as plt
from numpy import arange

# Pythonic modules
import os
import argparse
from subprocess import Popen, PIPE

# Own modules
from custom_functions import sh, cd, SmartFormatter
from meme_functions import passed

# A more permissive version of the motif reader found in the biopython module
# Take a look in the file for more details 
from motif_reader import read

def whoPassed(file):
    with open(file) as handle:
        record = read(handle)
    for gene in record:
        if passed(gene):
            print("[PASSED]: " + gene.consensus)
        else:
            print("[FAILED]: " + gene.consensus)

def postprocess(dire, verbose):
    candidates = list()
    included = list()
    genes = set() # read-only 
    covered = set()

    with cd(dire):
        print("postprocessing", dire)
        _, dirs, _ = next(os.walk('.'))
        meme_outs = filter(lambda d: d.startswith('meme_output'), dirs)
        for meme_out in meme_outs.splitlines():
            with cd(meme_out[2:], verbose=verbose):
                with open('meme.txt') as handle:
                    record = read(handle)
                
                genes = set(record.sequences) # after this, set for read-only
                candidates.extend([motif for motif in record if passed(motif)])
                
    # set up
    for motif in candidates:
        motif.occ = {instance.sequence_name for instance in motif.instances}
    
    while len(candidates) != 0:
        # include the best motif
        motif = max(candidates, key=lambda m: len(m.occ))
        covered |= motif.occ
        candidates.remove(motif)
        included.append(motif)
        # update the occurences of the motifs
        for motif in list(candidates): 
            motif.occ = set(motif.occ) - covered # only take acount 
                     # for coverage that we did not yet had
            if len(motif.occ) == 0:
                candidates.remove(motif)
    
    print("amount of motifs: ", len(included))
    print("amount of genes covered: ", len(covered))
    print("total amount of genes: ", len(genes))
    
    if len(genes) == 0:
        print("So the total percentage is: 0%")
    else:
        print("So the total percentage is:", "%.2f%%" % (len(covered)/len(genes) * 100))
        
    print(covered)
    
    coverage = list()
    for incl in included:
        if verbose: print("I cover", incl.occ)
        coverage.append(len(incl.occ))
    print(coverage)
    total = 0
    for index, c in enumerate(coverage):
        coverage[index] += total
        total += c
#    plt.close('all')
    
    if len(coverage) <= 1:
        print("[FAILED]: " + dire + " does not have enough motifs for a "
              + "meaningfull graph")
    else:
        print('PLOTTING ' + dire)
        coverage = [0] + coverage
        percentage = [c/len(genes) for c in coverage]
        motifs = range(len(percentage))
        plt.figure()
        plt.plot(motifs, percentage)
        plt.title('Coverage of genes of ' + dire)
        plt.xlabel('Amount of motifs')
        plt.ylabel('Percentage of genes covered')
    
    print()
    return len(genes)



def find(type):
    proc = Popen('find -type %s | wc -l' % type, stdout=PIPE, shell=True)
    if type == 'f':
        files = {file.split('/')[-1] for file in sh('find -name "*.fasta"').splitlines() 
                              if 'gathered' not in file}
        print(files)
        return len(files)
    for line in proc.stdout.readlines():
        return int(line.decode('ascii'))

def count(dire):
    types = ['d', 'f', 'd -empty'] # directories, files, empty directories
    dic = {}
    with cd(dire):
        for drc in os.listdir():
            with cd(drc):
                dic[drc] = [find(ar) for ar in types]
            
    for entry in dic.values():
        entry[0] -= (1 + entry[2]) 
        # from subcatergories remove empty files and it self
    return dic
    
#==============================================================================
# PLOTTING
#==============================================================================


def plot(dic):
    index = arange(len(dic))
    barwidth = 0.3
    plt.close('all')
    fig, ax = plt.subplots()
    xticks = ["".join(keys.split('_')[:-1]) for keys in dic.keys()]
    labels = ['Subcategories', 'Genes (Sequence known)', 'Genes (Sequence unkown)']
    for n in range(3):
        plt.bar(index + barwidth*n, 
                [entry[n] for entry in dic.values()], 
                barwidth, 
                label=labels[n])
    
    
    plt.ylabel('Amount')
    plt.xlabel('Category')
    plt.xticks(index + barwidth/2, xticks, rotation=45, ha='right')
    plt.title('Amount of genes per category')
    plt.legend()
    plt.tight_layout()
    
    plt.show()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="A program to calculate the "
                                     + "motifs that pass our requirements and "
                                     + "plot them in a meaningfull way", 
                                     formatter_class=SmartFormatter)
    parser.add_argument('-s', '--statistics', help="Give statistics of the "
                        + "given directory")
    parser.add_argument('-f', '--file', help="Give a file and calculate from "
                        + "this file the motifs that did pass")
    parser.add_argument('-d', '--dir', help="A directory to work in all "
                        + "subdirectories")
    parser.add_argument('-v', '--verbose', help="Execute the code in a more " 
                        + "verbose manner", action="store_true")
    parser.add_argument('dirs', help="List of dirs to work in the meme_output*", 
                        nargs=argparse.REMAINDER)
    args = parser.parse_args()
    if args.statistics:
        count(args.statistics)
    elif args.file:
        print("See who passed in " + args.file)
        whoPassed(args.file)
    elif args.dir:
        print("Postprocess in " + args.dir)
        with cd(args.dir):
            for dire in os.listdir():
                postprocess(args.dir, args.verbose)
    else:
        print("Postprocess in " + str(args.dirs))
        for dire in args.dir:
            postprocess(args.dir, args.verbose)
        
    

# TRASH BELOW        
#def work(dire):
#    candidates = list()
#    included = list()
#    genes = set() # read only 
#    covered = set()
#    
#    verbose = False
#    
##    with open("/home/hielke/py/antibiotic_molecule_36008/antibiotic_mixture_37087/meme_output_3/meme.txt") as handle:
##        record = read(handle)
##    for letter in record.alphabet.letters:
##        print(letter)
##    raise Exception
#    
#    print("post-processing", dire)
#    
#    with cd ('antibiotic_molecule_36008/' + dire, verbose=verbose):
#        meme_outs = sh('find . -maxdepth 1 -name "meme_output*"', verbose=verbose)
#        for meme_out in meme_outs.splitlines():
#            with cd (meme_out[2:], verbose=verbose):
#                with open('meme.txt') as handle:
#                    record = read(handle)
#                
#                genes = set(record.sequences)
#                candidates.extend([motif for motif in record if passed(motif)])
#                
#    
#    # set up
#    for motif in candidates:
#        motif.occ = {instance.sequence_name for instance in motif.instances}
#    
#    while len(candidates) != 0:
#        # include the best motif
#        motif = max(candidates, key=lambda m: len(m.occ))
#        covered |= motif.occ # update with union (on sets)
#        candidates.remove(motif)
#        included.append(motif)
#        # update the occurences of the motifs
#        for motif in list(candidates): 
#            motif.occ = set(motif.occ) - covered
#            if len(motif.occ) == 0:
#                candidates.remove(motif)
#    
#    print("amount of motifs: ", len(included))
#    print("amount of genes covered: ", len(covered))
#    print("total amount of genes: ", len(genes))
#    
#    if len(genes) == 0:
#        print("So the total percentage is: 0%")
#    else:
#        print("So the total percentage is:", "%.2f%%" % (len(covered)/len(genes) * 100))
#        
#    print(covered)
#    
#    
#    coverage = list()
#    for incl in included:
#        if verbose: print("I cover", incl.occ)
#        coverage.append(len(incl.occ))
#    print(coverage)
#    total = 0
#    for index, c in enumerate(coverage):
#        coverage[index] += total
#        total += c
##    plt.close('all')
#    
#    if len(coverage) <= 1:
#        print("[FAILED]: " + dire + " does not have enough motifs for a "
#              + "meaningfull graph")
#    else:
#        print('PLOTTING ' + dire)
#        coverage = [0] + coverage
#        percentage = [c/len(genes) for c in coverage]
#        motifs = range(len(percentage))
#        plt.figure()
#        plt.plot(motifs, percentage)
#        plt.title('Coverage of genes of ' + dire)
#        plt.xlabel('Amount of motifs')
#        plt.ylabel('Percentage of genes covered')
#    
#    print()
#    return len(genes)
#    
#    '''    
#    ['polyamine_antibiotic_36666' xxx
#     'peptide_antibiotic_36192' xxx
#     'macrolide_antibiotic_35919'  xxx
#     'lipopeptide_antibiotic_36175' xxx
#     'aminoglycoside_antibiotic_35935' xxx
#     'sulfonamide_antibiotic_36421' xxx
#     'rifamycin_antibiotic_36296' xxx
#     'acridine_dye_36193' xxx
#     'macrocyclic_antibiotic_37717' xxx
#     'antibiotic_mixture_37087' xxx
#     'pleuromutilin_antibiotic_37014' xxx
#     'tetracycline_35968' xxx
#     'antibacterial_free_fatty_acids_40727' xxx
#     'phenicol_antibiotic_36526' xxx
#     'elfamycin_antibiotic_37618' xxx
#     'fluoroquinolone_antibiotic_35920' xxx
#     'glycopeptide_antibiotic_36220' xxx
#     'diaminopyrimidine_antibiotic_36310' xxx
#     'beta-lactam_antibiotic_36016' x (only 20 percent)
#     'miscellaneous_antibiotic_36012' xxx
#     'aminocoumarin_antibiotic_36242' xxx
#     'streptogramin_antibiotic_35945' xxx
#     'nucleoside_antibiotic_36174' xxx
#     'oxazolidinone_antibiotic_36218' xxx
#     'tetracycline_derivative_36189'
#     'sulfone_39985' xxx
#     'lincosamide_antibiotic_35936' xxx
#     'organoarsenic_antibiotic_36690' xxx
#     ]"""      
#    '''
#    
#dirs = ['polyamine_antibiotic_36666', 
#     'peptide_antibiotic_36192' ,
#     'macrolide_antibiotic_35919',  
#     'lipopeptide_antibiotic_36175', 
#     'aminoglycoside_antibiotic_35935', 
#     'sulfonamide_antibiotic_36421' ,
#     'rifamycin_antibiotic_36296' ,
#     'acridine_dye_36193' ,
#     'macrocyclic_antibiotic_37717' ,
#     'antibiotic_mixture_37087' ,
#     'pleuromutilin_antibiotic_37014' ,
#     'tetracycline_35968' ,
#     'antibacterial_free_fatty_acids_40727' ,
#     'phenicol_antibiotic_36526' ,
#     'elfamycin_antibiotic_37618' ,
#     'fluoroquinolone_antibiotic_35920' ,
#     'glycopeptide_antibiotic_36220' ,
#     'diaminopyrimidine_antibiotic_36310' ,
#     'beta-lactam_antibiotic_36016' ,
#     'miscellaneous_antibiotic_36012', 
#     'aminocoumarin_antibiotic_36242' ,
#     'streptogramin_antibiotic_35945' ,
#     'nucleoside_antibiotic_36174' ,
#     'oxazolidinone_antibiotic_36218' ,
#     'tetracycline_derivative_36189',
#     'sulfone_39985' ,
#     'lincosamide_antibiotic_35936' ,
#     'organoarsenic_antibiotic_36690' 
#     ]
#
#print({dire: work(dire) for dire in dirs})
#
#'''
#EXTRA:
#    fluoroquinolone_antibiotic_35920
#    aminoglycoside_antibiotic_35935
#    
#'''
#
#if __name__ == '__main__':
#    file = '/home/hielke/py/antibiotic_molecule_36008/beta-lactam_antibiotic_36016/meme_output_250/meme.txt'
#    whoPassed(file)
