#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep  8 15:24:55 2017

@author: hielke
"""

# biopython, this has to be installed
from Bio import Entrez

# own modules
from custom_functions import cd, SmartFormatter

# pythonic modules
import re
from urllib.error import HTTPError
import argparse

def saveFasta(fastaString):
    filename = fastaString.strip().splitlines()[0]
    filename = re.split('\W+', filename)
    if not filename[0]:
        filename = filename[1:]
    filename = '_'.join(filename) + '.fasta'
    with open(filename, 'w') as f:
        f.write(fastaString)
        
def search(gene, retmax, orgas, mail):
    Entrez.email = mail
    term = gene + '[Gene]'
    if orgas:
        orgas = ['"' + orga + '"[Organism]' for orga in orgas]
        orgas = ' OR '.join(orgas)
        term += ' AND (' + orgas + ')'
    term += 'AND (alive[prop] OR replaced[Properties] OR discontinued[Properties])'
    print('Entrez.esearch: ' + term)
    return Entrez.esearch(db="gene", term=term, retmax=retmax)

def nucleotide(geneId, mail):
    Entrez.email = mail
    handle = Entrez.efetch(db="gene", id=geneId, rettype="gb", retmode="fasta")
    entry = handle.read()
    handle.close()
    print(entry)
    dc = {"db": "nuccore", "rettype": "fasta", "retmode": "text"}
    for line in entry.splitlines():
        if line.startswith("Annotation:"):
            line_split = line.split()
            
            seq = filter(bool, re.split('\D+', line_split[2]))
            try:
                dc.update({"id": line_split[1], 
                        "seq_start": next(seq),
                        "seq_stop": next(seq),
                        # If len is 4, we have a complement.
                        "strand": 2 if len(line_split) == 4 else 1
                        })
            except StopIteration:
                print("unable to parse", line)
                return ""
    print(dc)
    try:
        handle = Entrez.efetch(**dc)
        fasta = handle.read()
        handle.close()
    except HTTPError:
        print("Unable to retrieve, either bad internetconnection, or a removed entry: ", geneId)
        return ""
    return fasta

def init(gene, orgas, mail):
    handle = search(gene, 0, orgas, mail)
    count = Entrez.read(handle)["Count"]
    handle.close()
    handle = search(gene, count, orgas, mail)
    record = Entrez.read(handle)
    handle.close()
    
    for Id in record['IdList']:
        fasta = nucleotide(Id)
        if not fasta:
            continue
        with cd(gene):
            saveFasta(fasta)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="A script to retrieve all genes "
                                     + "from the NCBI database, and save them. "
                                     + "It also is able to retrieve genes that "
                                     + "have no own entry but are only annotated.",
                                     formatter_class=SmartFormatter)
    parser.add_argument('-m', '--mail', help="You have to give your email address "
                        + "in order to be allowed to interact with the NCBI API.")
    parser.add_argument('-g', '--gene', help="The gene from which you want to "
                        + "retrieve all the variants.")
    parser.add_argument('orgas', help="At the end you can optional specify "
                        + "the organisms you want to filter on", 
                        nargs=argparse.REMAINDER)
    args = parser.parse_args()
    
    if not args.mail:
        print("You have to specify a email address in order to be allowed to "
              + "interact with the NCBI API.")
    elif not args.gene:
        print("You have to give a gene to search for.")
    else:
        init(args.gene, args.orgas, args.mail)

#if False:
#    
#    handle = Entrez.esearch(db="gene", term="blaZ[Gene Name] AND alive[prop]", retmax=0)
#    record = Entrez.read(handle)
#    handle.close()
#    count = record["Count"]
#    idd = '2598287'
#    
#    
#    handle = Entrez.efetch(db="gene", id=idd, rettype="gb", retmode="fasta")
#    entry = handle.read()
#    handle.close()
#    dc = {"db": "nuccore", "rettype": "fasta", "retmode": "text"}
#    for line in entry.splitlines():
#        if line.startswith("Annotation:"):
#            line_split = line.split()
#            seq = filter(bool, re.split('\D+', line_split[2]))
#            dc.update({"id": line_split[1], 
#                    "seq_start": next(seq),
#                    "seq_stop": next(seq),
#                    "strand": 2 if line_split[3].startswith('complement') else 1
#                    })
#    print(dc)
#    handle = Entrez.efetch(**dc)
#    print(handle.read())
#    handle.close()
#    
#    
#    
    #
    #A = """
    #1. blaZ
    #beta-lactamase [Staphylococcus aureus subsp. aureus N315]
    #Other Aliases: SAP010
    #Genomic context: Plasmid pN315
    #Annotation:  NC_003140.1 (7945..8790, complement)
    #This record was discontinued.
    #ID: 1122765
    #"""
    #def tryKwargs(id, seq_stop, seq_start, strand):
    #    print('id', id)
    #    print('seq_stop', seq_stop)
    #    print('seq_start', seq_start)
    #    print('strand', strand)
    #    
    #for line in A.splitlines():
    #    if line.startswith("Annotation:"):
    #        line_split = line.split()
    #        seq = filter(bool, re.split('\D+', line_split[2]))
    #        dc = {"id": line_split[1], 
    #                "seq_start": next(seq),
    #                "seq_stop": next(seq),
    #                "strand": 2 if line_split[3].startswith('complement') else 1
    #                }
    #        print(dc)
    #        tryKwargs(**dc)
    #        
